from attrs import define

from kubekind.kind import Kind

from .container import Container


@define
class Pod(Kind):
    """
    Pod is a collection of containers that can run on a host.
    In a Pod you can define objects of type `Container` and `Volumes`

    Args:
        name: Name of the pod

    Details: https://kubernetes.io/docs/reference/kubernetes-api/workload-resources/pod-v1/"""

    name: str
    apiVersion = "v1"
    allowed_classes = [Container]



from attrs import define

from kubekind.kind import RawObject


@define
class Container(RawObject):
    """
    A single application container that you want to run within a pod.

    Args:
        name: Name of the container specified as a DNS_LABEL
        image: Container image name
        args: Arguments to the entrypoint. The container image's CMD is used if this is not provided.

    Details: https://kubernetes.io/docs/reference/kubernetes-api/workload-resources/pod-v1/#Container
    """

    name: str
    image: str
    args: list[str] = []
    prefix_key = "containers"



from collections import defaultdict
from typing import Literal

import yaml
from attrs import NOTHING, define, fields
from rich import print as rich_print
from rich import print_json
from rich.syntax import Syntax


class NotAllowedExecption(Exception):
    pass


class Stack:
    """
    Stack singleton to keep track of context managers
    Required the add(None) method
    """

    stack = []

    @classmethod
    def add(cls, item: object):
        return Stack.stack.append(item)

    @classmethod
    def pop(cls):
        return Stack.stack.pop()

    @classmethod
    def tail(cls):
        return Stack.stack[-1]


class RawObject:
    def __init__(self):
        self.obj = {}
        self.childs = []

    def print(self, mode: Literal[" yaml", "json"] = "yaml"):
        data = self.as_dict()
        if mode == "json":
            print_json(data=data)
        else:
            yaml_output = Syntax(yaml.dump(data, sort_keys=False), "yaml")
            rich_print(yaml_output)

    def __enter__(self):
        if Stack.stack:
            parent = self.add()
        else:
            Stack.add(self)
            parent = None
        self.parent = parent
        return self

    def __exit__(self, type, value, traceback):
        Stack.pop()

    def add(self, obj: object = None):
        """
        Add object

        If is None add 'self' to the parent context manager object
        Else add 'obj' as chidl to 'self'
        """
        if obj is None:
            parent = Stack.tail()
            parent.check_is_allowed_class(self)
            parent.childs.append(self)
            return self
        else:
            self.check_is_allowed_class(obj)
            self.childs.append(obj)
            obj.parent = self
            return obj

    def check_is_allowed_class(self, obj: object):
        allowed_classes = getattr(self, "allowed_classes")
        for cls in allowed_classes:
            if isinstance(obj, cls):
                return True
        raise NotAllowedExecption(
            f"'{obj.__class__.__name__}' not allowed in '{self.__class__.__name__}'"
        )

    @property
    def __members__(self):
        member_dict = {}
        for name in dir(self):
            if "__" not in name:
                member_dict[name] = getattr(self, name)
        return member_dict

    def as_dict(self) -> dict:
        members = self.__members__
        expose_dict = {}
        for field in fields(self.__class__):
            value = members[field.name]
            if field.default is NOTHING or value:
                expose_dict[field.name] = value
        return expose_dict


@define
class Kind(RawObject):
    def __attrs_pre_init__(self):
        super().__init__()

    def as_dict(self):
        obj = {
            "kind": self.__class__.__name__,
            "apiVersion": self.apiVersion,
            "metadata": {"name": self.name},
        }
        spec_classes = getattr(self, "allowed_classes", [])
        extra_dict = defaultdict(list)
        for child in self.childs:
            if child.__class__ in spec_classes:
                extra_dict[child.prefix_key].append(child.as_dict())
        spec_dict = {}
        if extra_dict:
            spec_dict = {"spec": dict(extra_dict)}
        return dict(obj, **spec_dict)



import pytest
from test_resources_def import (
    CONTAINER_EXPECTED_RESULT,
    POD_CONTAINER_EXPECTED_RESULT,
    POD_EXPECTED_RESULT,
)

from kubekind.resources.container import Container
from kubekind.resources.pod import Pod


def test_pod_missing_parameter():
    with pytest.raises(TypeError) as _:
        with Pod():
            pass


def test_pod_def():

    with Pod("my-pod") as pod:
        assert pod.as_dict() == POD_EXPECTED_RESULT
        container = Container("bash", "bash:latest").add()
        assert container.as_dict() == CONTAINER_EXPECTED_RESULT
    assert pod.as_dict() == POD_CONTAINER_EXPECTED_RESULT
    pod.print("json")
    pod.print("yaml")



import pytest

from kubekind.resources.container import Container
from kubekind.resources.pod import Pod

POD_EXPECTED_RESULT: dict = {
    "kind": "Pod",
    "apiVersion": "v1",
    "metadata": {"name": "my-pod"},
}

CONTAINER_EXPECTED_RESULT: dict = {"name": "bash", "image": "bash:latest"}

POD_CONTAINER_EXPECTED_RESULT = {
    **POD_EXPECTED_RESULT,
    **{"spec": {"containers": [CONTAINER_EXPECTED_RESULT]}},
}


def test_pod_missing_parameter():
    with pytest.raises(TypeError) as _:
        Pod()


def test_pod_def():

    pod = Pod("my-pod")
    assert pod.as_dict() == POD_EXPECTED_RESULT

    container = Container("bash", "bash:latest")
    assert container.as_dict() == CONTAINER_EXPECTED_RESULT

    pod.add(container)
    assert pod.as_dict() == POD_CONTAINER_EXPECTED_RESULT



from kubekind import Container, Pod

with Pod("my-pod") as pod:
    Container("bash", "bash:latest", args=["sleep", "10000"]).add()
    Container("bash_2", "bash:latest", args=["sleep", "10000"]).add()

pod.print()
